/***************************************************************************/
/*                                                                         */
/*  ftsystem.c                                                             */
/*                                                                         */
/*    ANSI-specific FreeType low-level system interface (body).            */
/*                                                                         */
/*  Copyright 1996-2001, 2002 by                                           */
/*  David Turner, Robert Wilhelm, and Werner Lemberg.                      */
/*                                                                         */
/*  This file is part of the FreeType project, and may only be used,       */
/*  modified, and distributed under the terms of the FreeType project      */
/*  license, LICENSE.TXT.  By continuing to use, modify, or distribute     */
/*  this file you indicate that you have read the license and              */
/*  understand and accept it fully.                                        */
/*                                                                         */
/***************************************************************************/

#include <pandora/main.h>
#include <pandora/files/file.h>

extern struct KernelBase *KernelBase;

#include <ft2build.h>
#include FT_CONFIG_CONFIG_H
#include FT_INTERNAL_DEBUG_H
#include FT_SYSTEM_H
#include FT_ERRORS_H
#include FT_TYPES_H

/*************************************************************************/

FT_CALLBACK_DEF( void* ) ft_alloc(FT_Memory memory, long size )
{
   APTR address;
   
   FT_UNUSED(memory);

   if (AllocMemory(size, MEM_DATA|MEM_UNTRACKED, &address, NULL) IS ERR_Okay) {
      return(address);
   }
   else return(NULL);
}

/*************************************************************************/

FT_CALLBACK_DEF(void *) ft_realloc( FT_Memory memory, long cur_size, long new_size, void *block)
{
   FT_UNUSED( memory );
   FT_UNUSED( cur_size );

   if (ReallocMemory(block, new_size, &block, NULL) IS ERR_Okay) {
      return(block);
   }
   else return(NULL);
}

/*************************************************************************/

FT_CALLBACK_DEF(void) ft_free(FT_Memory  memory, void *block)
{
   FT_UNUSED(memory);

   FreeMemory(block);
}

/*************************************************************************/
/*                                                                       */
/* The macro FT_COMPONENT is used in trace mode.  It is an implicit      */
/* parameter of the FT_TRACE() and FT_ERROR() macros, used to print/log  */
/* messages during execution.                                            */
/*                                                                       */

#undef  FT_COMPONENT
#define FT_COMPONENT  trace_io

  /* We use the macro STREAM_FILE for convenience to extract the       */
  /* system-specific stream handle from a given FreeType stream object */
#define STREAM_FILE( stream )  ( (FILE*)stream->descriptor.pointer )

  /*************************************************************************/
  /* <Function>                                                            */
  /*    ft_ansi_stream_close                                               */

FT_CALLBACK_DEF( void ) ft_ansi_stream_close( FT_Stream  stream )
{
   Action(AC_Free, STREAM_FILE(stream), NULL);

   stream->descriptor.pointer = NULL;
   stream->size               = 0;
   stream->base               = 0;
}

  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    ft_ansi_stream_io                                                  */
  /*                                                                       */
  /* <Description>                                                         */
  /*    The function to open a stream.                                     */
  /*                                                                       */
  /* <Input>                                                               */
  /*    stream :: A pointer to the stream object.                          */
  /*                                                                       */
  /*    offset :: The position in the data stream to start reading.        */
  /*                                                                       */
  /*    buffer :: The address of buffer to store the read data.            */
  /*                                                                       */
  /*    count  :: The number of bytes to read from the stream.             */
  /*                                                                       */
  /* <Return>                                                              */
  /*    The number of bytes actually read.                                 */
  /*                                                                       */

FT_CALLBACK_DEF(unsigned long) ft_ansi_stream_io(FT_Stream stream, unsigned long offset, unsigned char *buffer, unsigned long count)
{
   struct acRead read;
   OBJECTPTR file;

   file = (OBJECTPTR)STREAM_FILE(stream);

   ActionTags(AC_Seek, file, (DOUBLE)offset, POS_START);

   if (count < 1) return(0);
   
   read.Buffer = buffer;
   read.Length = count;
   if (Action(AC_Read, file, &read) IS ERR_Okay) {
      return(read.Result);
   }
   else return(0);
}

/****************************************************************************/

FT_EXPORT_DEF(FT_Error) FT_Stream_Open( FT_Stream stream, const char*  filepathname )
{
   OBJECTPTR file;

   if (!stream) return FT_Err_Invalid_Stream_Handle;

   if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
      FID_Location|TSTRING, filepathname,
      FID_Flags|TLONG,      FL_READ,
      TAGEND) IS ERR_Okay) {

      GetField(file, FID_Size, FT_LONG, &stream->size);

      stream->descriptor.pointer = file;
      stream->pathname.pointer   = (char *)filepathname;
      stream->pos                = 0;
      stream->read  = ft_ansi_stream_io;
      stream->close = ft_ansi_stream_close;
      return(FT_Err_Ok);
   }
   else return FT_Err_Cannot_Open_Resource;
}

/****************************************************************************/

FT_EXPORT_DEF(FT_Memory) FT_New_Memory(void)
{
   FT_Memory  memory;
   APTR address;
 
   if (AllocMemory(sizeof(*memory), MEM_DATA|MEM_UNTRACKED, &address, NULL) IS ERR_Okay) {
      memory = address;
      memory->user    = 0;
      memory->alloc   = ft_alloc;
      memory->realloc = ft_realloc;
      memory->free    = ft_free;
      return(memory);
   }
   else return(NULL);
}

/****************************************************************************/

FT_EXPORT_DEF( void ) FT_Done_Memory( FT_Memory  memory )
{
   FreeMemory(memory);
}
