/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Resize
Version:   1.1
ID:        ID_RESIZE
Category:  GUI
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2004.  All rights reserved.
Keywords:  resize, expand, shrink, enlarge, scale
Short:     The Resize class controls the resizing of rendered areas.

###DESCRIPTION###
<p>The Resize class is used for creating user-interactive resizing areas.  In
most cases it is applied to the edges of Render objects so that the user can
make simple adjustments to display areas.  When creating a new Resize object, you can
choose what edges of the drawable border should be monitored for resizing, or
alternatively you may pin-point the resizing area through standard dimension
specifications.  The following example demonstrates the use of both methods:</p>

<pre>
  &lt;render x=&quot;50&quot; y=&quot;70&quot; width=&quot;250&quot; height=&quot;300&quot;&gt;
    &lt;resize border=&quot;left|right|top|bottom&quot; bordersize=&quot;10&quot;/&gt;

    &lt;resize xoffset=&quot;10&quot; yoffset=&quot;10&quot; width=&quot;20&quot; height=&quot;20&quot; direction=&quot;all&quot;/&gt;
  &lt;/render&gt;
</pre>

<p>The first Resize object monitors all four sides of the rendered area, to a
region not exceeding 10 units on either edge.  To do this, we simply specified
the borders that are to be monitored through the Border field.  The second
Resize object monitors an area that is 20x20 units in size at an offset of 10
units from the bottom right edge.  The Direction field has been set to a value
of 'all', which means that the user can resize the drawable area in any
direction by interacting with the Resize object.</p>

<p>When using Resize objects to manage the dimensions of rendered areas, it
is recommended that the MinWidth, MinHeight, MaxWidth and MaxHeight fields
are used to prevent excessive shrinkage or expansion.  These values must be
set in the Render object that the resize functionality is being applied to.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/input/joyport.h>
#include <pandora/graphics/all.h>
#include <pandora/modules/pointer.h>
#include <pandora/modules/render.h>

#undef NULL
#define NULL 0

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_RESIZE 1.1

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_RESIZE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004.  All rights reserved.",
   "January 2004",
   "Resize"
};

static struct KernelBase *KernelBase;
static struct RenderBase *RenderBase;
static struct PointerBase *PointerBase;
static OBJECTPTR modRender = NULL, modPointer = NULL;
static OBJECTPTR ResizeClass = NULL;
static OBJECTID glPointerID = 0;

/*****************************************************************************
** Class definition.
*/

struct Resize {
   OBJECT_HEADER          /* Class header */
   OBJECTID Object;       /* Object that is to be resized */
   OBJECTID DrawableID;   /* Drawable object that is to be used for monitoring */
   LONG     Button;       /* Determines what button is used for resizing */
   LONG     Direction;    /* Direction flags (horizontal/vertical) */
   LONG     BorderFlags;  /* Border flags can be used to monitor up to 8 seperate areas at once */
   LONG     BorderSize;   /* Determines the size of the border edge */

   /*** Private fields ***/
   WORD     State;
   LONG     PointerX;
   LONG     PointerY;
   LONG     OriginalWidth;
   LONG     OriginalHeight;
   LONG     OriginalX;
   LONG     OriginalY;
   LONG     Dimensions;
   FLOAT    XOffset;
   FLOAT    YOffset;
   FLOAT    XCoord;         /* X coordinate of the monitored region */
   FLOAT    YCoord;         /* Y coordinate of the monitored region */
   FLOAT    Width;          /* Width of the monitored region */
   FLOAT    Height;         /* Height of the monitored region */
   BYTE     prvAnchored;
   WORD     prvAnchorX;
   WORD     prvAnchorY;
   WORD     PointerLocked;
};

static struct FieldDef DirectionFlags[] = {
   { "DOWN",  MOVE_DOWN }, { "UP",    MOVE_UP    },
   { "LEFT",  MOVE_LEFT }, { "RIGHT", MOVE_RIGHT },
   { "ALL",   MOVE_ALL  },
   { NULL, NULL }
};

static struct FieldDef BorderFlags[] = {
   { "TOP",         EDGE_TOP },
   { "LEFT",        EDGE_LEFT },
   { "RIGHT",       EDGE_RIGHT },
   { "BOTTOM",      EDGE_BOTTOM },
   { "TOPLEFT",     EDGE_TOPLEFT },
   { "TOPRIGHT",    EDGE_TOPRIGHT },
   { "BOTTOMLEFT",  EDGE_BOTTOMLEFT },
   { "BOTTOMRIGHT", EDGE_BOTTOMRIGHT },
   { "ALL",         EDGE_ALL },
   { NULL, NULL }
};

/*** Field Prototypes ***/

static ERROR GET_Height(struct Resize *, struct Variable *);
static ERROR GET_Width(struct Resize *, struct Variable *);
static ERROR GET_X(struct Resize *, struct Variable *);
static ERROR GET_XOffset(struct Resize *, struct Variable *);
static ERROR GET_Y(struct Resize *, struct Variable *);
static ERROR GET_YOffset(struct Resize *, struct Variable *);

static ERROR SET_BorderSize(struct Resize *, LONG *);
static ERROR SET_Height(struct Resize *, struct Variable *);
static ERROR SET_Width(struct Resize *, struct Variable *);
static ERROR SET_X(struct Resize *, struct Variable *);
static ERROR SET_XOffset(struct Resize *, struct Variable *);
static ERROR SET_Y(struct Resize *, struct Variable *);
static ERROR SET_YOffset(struct Resize *, struct Variable *);

/*** Class Definition ***/

static struct FieldArray ResizeFields[] = {
   { "Object",     0, FDF_OBJECTID|FDF_RW,  0,                     NULL, NULL },
   { "Drawable",   0, FDF_OBJECTID|FDF_RW,  ID_RENDER,             NULL, NULL },
   { "Button",     0, FDF_LONG|FDF_RW,      0,                     NULL, NULL },
   { "Direction",  0, FDF_LONGFLAGS|FDF_RW, (LONG)&DirectionFlags, NULL, NULL },
   { "Border",     0, FDF_LONGFLAGS|FDF_RW, (LONG)&BorderFlags,    NULL, NULL },
   { "BorderSize", 0, FDF_LONG|FDF_RW,     0,                     NULL, SET_BorderSize },
   /*** Variable Fields ***/
   { "Height",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",          0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",    0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",          0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",    0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*** Action Prototypes ***/

static ERROR RESIZE_ActionNotify(struct Resize *, struct acActionNotify *);
static ERROR RESIZE_ClosingTag(struct Resize *, APTR);
static ERROR RESIZE_Free(struct Resize *, APTR);
static ERROR RESIZE_Init(struct Resize *, APTR);
static ERROR RESIZE_NewObject(struct Resize *, APTR);

static struct ActionArray ResizeActions[] = {
   { AC_ActionNotify, RESIZE_ActionNotify },
   { AC_ClosingTag,   RESIZE_ClosingTag },
   { AC_Free,         RESIZE_Free },
   { AC_Init,         RESIZE_Init },
   { AC_NewObject,    RESIZE_NewObject },
   { NULL, NULL }
};

/****************************************************************************/

static LONG WithinArea(struct Resize *, LONG, LONG);
static ERROR GetSystemPointer(struct Pointer **);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modRender, NULL,
      FID_Name|TSTRING,   "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modPointer, NULL,
      FID_Name|TSTRING,   "pointer",
      FID_Version|TFLOAT, MODVERSION_POINTER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modPointer, FID_ModBase, FT_POINTER, &PointerBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   FastFindObject("SystemPointer", ID_POINTER, &glPointerID, 1, NULL);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ResizeClass, NULL,
      FID_BaseClassID|TLONG, ID_RESIZE,
      FID_Name|TSTRING,      "Resize",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      ResizeActions,
      FID_Fields|TPTR,       ResizeFields,
      FID_Size|TLONG,        sizeof(struct Resize),
      FID_Version|TFLOAT,    VER_RESIZE,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (ResizeClass) { acFree(ResizeClass); ResizeClass = NULL; }
   if (modRender)   { acFree(modRender);   modRender = NULL; }
   if (modPointer)  { acFree(modPointer);  modPointer = NULL; }
   return(ERR_Okay);
}

static LONG GetCursorType(struct Resize *Self)
{
   switch (Self->Direction) {
      case (MOVE_UP|MOVE_LEFT):    return(PTR_SIZE_TOP_LEFT); 
      case (MOVE_UP|MOVE_RIGHT):   return(PTR_SIZE_TOP_RIGHT);
      case (MOVE_DOWN|MOVE_LEFT):  return(PTR_SIZE_BOTTOM_LEFT); 
      case (MOVE_DOWN|MOVE_RIGHT): return(PTR_SIZE_BOTTOM_RIGHT);
      case MOVE_LEFT:              return(PTR_SIZE_LEFT); 
      case MOVE_RIGHT:             return(PTR_SIZE_RIGHT); 
      case MOVE_UP:                return(PTR_SIZE_TOP); 
      case MOVE_DOWN:              return(PTR_SIZE_BOTTOM); 
      default:                     return(PTR_SIZING); 
   }
}

/*****************************************************************************
** Resize: ActionNotify
*/

#define CLICK_HELD     1
#define CLICK_RELEASED 0

static ERROR RESIZE_ActionNotify(struct Resize *Self, struct acActionNotify *NotifyArgs)
{
   if (NotifyArgs->Error != ERR_Okay) return(ERR_Okay);

   if (NotifyArgs->ActionID IS AC_UserMovement) {
      struct acUserMovement *movement;
      struct Pointer *pointer;

      movement = (struct acUserMovement *)NotifyArgs->Args;

      if (Self->State IS CLICK_HELD) {
         struct acRedimension redimension;
         LONG px, py, absx, absy;
         OBJECTPTR object;
         OBJECTID pointeranchor;
         ERROR error;

         if ((error = GetSystemPointer(&pointer)) IS ERR_Okay) {
            if ((pointeranchor = pointer->AnchorID)) {
               px = pointer->XCoord + movement->XChange;
               py = pointer->YCoord + movement->YChange;
            }
            else {
               px = pointer->XCoord;
               py = pointer->YCoord;
            }
            ReleaseObject(pointer);
         }
         else {
            DPrintF("@Resize:","Failed to access SystemPointer, error %d.", error);
            return(error);
         }

         if ((error = AccessObject(Self->Object, 4000, &object) IS ERR_Okay)) {
            /*** Send the resize message to the target object ***/

            LONG minwidth, maxwidth, minheight, maxheight, left, right, top, bottom;

            GetFields(object, FID_MinWidth|TLONG,     &minwidth,
                              FID_MaxWidth|TLONG,     &maxwidth,
                              FID_MinHeight|TLONG,    &minheight,
                              FID_MaxHeight|TLONG,    &maxheight,
                              FID_LeftMargin|TLONG,   &left,
                              FID_TopMargin|TLONG,    &top,
                              FID_RightMargin|TLONG,  &right,
                              FID_BottomMargin|TLONG, &bottom,
                              TAGEND);

            maxwidth += left + right;
            maxheight += top + bottom;
            minwidth += left + right;
            minheight += top + bottom;

            if (Self->Direction & MOVE_RIGHT) {
               /*** Movement comes from the right edge of the drawable ***/
               redimension.XCoord = Self->OriginalX;
               redimension.Width = Self->OriginalWidth + (px - Self->PointerX);
            }
            else if (Self->Direction & MOVE_LEFT) {
               /*** Movement comes from the left edge of the drawable ***/
               if (px >= Self->OriginalX + Self->OriginalWidth) return(ERR_Okay);
               redimension.XCoord = Self->OriginalX - (Self->PointerX - px);
               redimension.Width  = Self->OriginalWidth - (px - Self->PointerX);

               /*** Check minwidth/maxwidth settings due to 'reverse resizing' ***/

               if (redimension.Width > maxwidth) {
                  redimension.XCoord = Self->OriginalX + Self->OriginalWidth - maxwidth;
                  redimension.Width = maxwidth;
               }
               else if (redimension.Width < minwidth) {
                  redimension.XCoord = Self->OriginalX + Self->OriginalWidth - minwidth;
                  redimension.Width  = minwidth;
               }
            }
            else {
               redimension.XCoord = Self->OriginalX;
               redimension.Width  = 0;
            }

            if (Self->Direction & MOVE_DOWN) {
               redimension.YCoord = Self->OriginalY;
               redimension.Height = Self->OriginalHeight + (py - Self->PointerY);
            }
            else if (Self->Direction & MOVE_UP) {
               redimension.YCoord = Self->OriginalY - (Self->PointerY - py);
               redimension.Height = Self->OriginalHeight - (py - Self->PointerY);

               /*** Check minheight/maxheight settings due to 'reverse resizing' ***/

               if (redimension.Height > maxheight) {
                  redimension.YCoord = Self->OriginalY + Self->OriginalHeight - maxheight;
                  redimension.Height = maxheight;
               }
               else if (redimension.Height < minheight) {
                  redimension.YCoord = Self->OriginalY + Self->OriginalHeight - minheight;
                  redimension.Height = minheight;
               }
            }
            else {
               redimension.YCoord = Self->OriginalY;
               redimension.Height = 0;
            }

            if (redimension.Width < 0)  redimension.Width  = 0;
            if (redimension.Height < 0) redimension.Height = 0;
            redimension.ZCoord = 0;
            redimension.Depth  = 0;

            Action(AC_Redimension, object, &redimension);
            //DelayMsg(AC_Redimension, object->UniqueID, &redimension); //<- Only works if anchoring is disabled.

            /* If we have anchored the pointer, we need to tell the pointer
            ** to move or else it will stay locked at its current position.
            */

            if (Self->prvAnchored) {
               GetFields(object, FID_Width|TDOUBLE,  &redimension.Width,
                                 FID_Height|TDOUBLE, &redimension.Height,
                                 TAGEND);

               ReleaseObject(object);

               if (pointeranchor IS Self->DrawableID) {
                  if (AccessObject(Self->DrawableID, 3000, &object) IS ERR_Okay) {
                     GetFields(object, FID_AbsX|TLONG, &absx,
                                       FID_AbsY|TLONG, &absy,
                                       TAGEND);
                     ReleaseObject(object);

                     if (Self->Direction & MOVE_RIGHT) {
                        absx = ((absx + redimension.Width) - (Self->OriginalWidth - Self->prvAnchorX));
                     }
                     else absx = (absx + Self->prvAnchorX);

                     if (Self->Direction & MOVE_DOWN) {
                        absy = ((absy + redimension.Height) - (Self->OriginalHeight - Self->prvAnchorY));
                     }
                     else absy = (absy + Self->prvAnchorY);

                     SetCursorPos(absx, absy);
                  }
                  else DPrintF("@Resize:","Unable to access anchored drawable #%d.", Self->DrawableID);
               }
               else Self->prvAnchored = FALSE;
            }
            else ReleaseObject(object);
         }
         else if (error IS ERR_NoMatchingObject) {
            Self->Object = NULL;
            Action(AC_Free, Self, NULL); /* Commit suicide */
         }
      }
      else {
         /* If the user is moving the mouse pointer over the resizing area
         ** and the mouse button is not currently held, check if we can
         ** change the pointer image to something else.  This provides
         ** effective visual notification to the user.
         */

         if (movement->OverObject IS Self->DrawableID) {
            LONG cursor, x, y;
            
            x = movement->XCoord;
            y = movement->YCoord;
            GetRelativeCursorPos(Self->DrawableID, &x, &y);

            if (WithinArea(Self, x, y)) {
               /*** Determine what cursor we should be using ***/

               cursor = GetCursorType(Self);
               
               /* If the cursor is to change, use SetCursor() to do it.  Note
               ** that ACF_ANCHOR is used to nullify existing anchorage settings.
               */

               if (cursor != Self->PointerLocked) {
                  if (SetCursor(NULL, ACF_BUFFER|ACF_ANCHOR|ACF_NOBUTTONS, cursor, NULL, Self->Head.UniqueID, NULL) IS ERR_Okay) {
                     Self->PointerLocked = cursor;
                  }
               }
            }
            else if (Self->PointerLocked) {
               RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
               Self->PointerLocked = NULL;
            }
         }
         else if (Self->PointerLocked) {
            RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
            Self->PointerLocked = NULL;
         }
      }

      return(ERR_Okay);
   }
   else if ((NotifyArgs->ActionID IS AC_UserClick) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct acUserClick *click;
      OBJECTPTR object;

      click = (struct acUserClick *)NotifyArgs->Args;

      /*** Check the correct button has been pressed ***/

      if (Self->Button IS 1) {
         if (!(click->ButtonFlags & JD_LMB)) return(ERR_Okay);
      }
      else if (Self->Button IS 2) {
         if (!(click->ButtonFlags & JD_RMB)) return(ERR_Okay);
      }
      else if (Self->Button IS 3) {
         if (!(click->ButtonFlags & JD_MMB)) return(ERR_Okay);
      }
      else {
         DPrintF("@Resize:","Incorrect button setting.");
         Action(AC_Free, Self, NULL);
         return(ERR_Failed);
      }

      /* Check the region to make sure that the button click
      ** has fallen in the correct place.
      */
  
      if (WithinArea(Self, click->XCoord, click->YCoord)) {
         if (AccessObject(Self->Object, 3000, &object) IS ERR_Okay) {
            GetFields(object, FID_XCoord|TLONG, &Self->OriginalX,
                              FID_YCoord|TLONG, &Self->OriginalY,
                              FID_Width|TLONG,  &Self->OriginalWidth,
                              FID_Height|TLONG, &Self->OriginalHeight,
                              TAGEND);
            ReleaseObject(object);

            /*** Anchor the pointer ***/

            if (GetCursorPos(&Self->PointerX, &Self->PointerY) IS ERR_Okay) {
               if (SetCursor(Self->DrawableID, ACF_ANCHOR, 0, NULL, Self->Head.UniqueID, NULL) IS ERR_Okay) {
                  Self->prvAnchored = TRUE;
                  Self->prvAnchorX  = click->XCoord;
                  Self->prvAnchorY  = click->YCoord;
               }

               Self->State = CLICK_HELD;
            }
            else ObjectError(Self, ERH_ActionNotify, ERR_Failed);
         }
         else DPrintF("@UserClick:","[Resize] Faied to access drawable #%d.", Self->Object);
      }

      return(ERR_Okay);
   }
   else if ((NotifyArgs->ActionID IS AC_UserClickRelease) AND (Self->State IS CLICK_HELD)) {
      struct acUserClickRelease *release;
      LONG x, y;

      release = (struct acUserClickRelease *)NotifyArgs->Args;

      /*** Click Release Handling ***/

      if (Self->Button IS 1) {
         if (!(release->ButtonFlags & JD_LMB)) return(ERR_Okay);
      }
      else if (Self->Button IS 2) {
         if (!(release->ButtonFlags & JD_RMB)) return(ERR_Okay);
      }
      else if (Self->Button IS 3) {
         if (!(release->ButtonFlags & JD_MMB)) return(ERR_Okay);
      }
      else {
         DPrintF("@Resize:","Incorrect button setting.");
         Action(AC_Free, Self, NULL);
         return(ERR_Failed);
      }

      if ((GetRelativeCursorPos(Self->DrawableID, &x, &y) IS ERR_Okay) AND
          (WithinArea(Self, x, y))) {

         /* Do not change the cursor image, just release the anchor.
         ** We will continue to retain our ownership of the cursor until it
         ** leaves our resize area.
         */

         SetCursor(NULL, ACF_ANCHOR|ACF_RESTRICT, GetCursorType(Self), NULL, Self->Head.UniqueID, NULL);
      }
      else {
         /*** Release the pointer ***/

         if (Self->PointerLocked) {
            RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
            Self->PointerLocked = NULL;
         }
      }

      Self->State = CLICK_RELEASED;

      return(ERR_Okay);
   }

   return(ERR_Failed);
}

/******************************************************************************
** Internal: WithinArea()
** Short:    Checks if the given coordinates fall within the given area.
*/

static LONG WithinArea(struct Resize *Self, LONG AreaX, LONG AreaY)
{
   struct Variable var;
   struct RenderInfo info;
   LONG x, y, width, height;

   if (Self->BorderFlags) {
      if (GetRenderInfo(Self->DrawableID, &info, sizeof(info)) != ERR_Okay) return(NULL);

      width = info.Width;
      height = info.Height;

      if (Self->BorderFlags & EDGE_TOPLEFT) {
         if ((AreaX >= 0) AND (AreaY >= 0) AND
             (AreaX < Self->BorderSize * 2) AND (AreaY < Self->BorderSize * 2)) {
            Self->Direction = MOVE_UP|MOVE_LEFT;
            return(EDGE_TOPLEFT);
         }
      }

      if (Self->BorderFlags & EDGE_TOPRIGHT) {
         if ((AreaX >= width - (Self->BorderSize * 2)) AND (AreaY >= 0) AND
             (AreaX < width) AND (AreaY < Self->BorderSize * 2)) {
            Self->Direction = MOVE_UP|MOVE_RIGHT;
            return(EDGE_TOPRIGHT);
         }
      }

      if (Self->BorderFlags & EDGE_BOTTOMLEFT) {
         if ((AreaX >= 0) AND (AreaY >= height - (Self->BorderSize*2)) AND
             (AreaX < Self->BorderSize*2) AND (AreaY < height)) {
            Self->Direction = MOVE_DOWN|MOVE_LEFT;
            return(EDGE_BOTTOMLEFT);
         }
      }

      if (Self->BorderFlags & EDGE_BOTTOMRIGHT) {
         if ((AreaX >= width - (Self->BorderSize*2)) AND (AreaY >= height - (Self->BorderSize*2)) AND
             (AreaX < width) AND (AreaY < height)) {
            Self->Direction = MOVE_DOWN|MOVE_RIGHT;
            return(EDGE_BOTTOMRIGHT);
         }
      }

      if (Self->BorderFlags & EDGE_TOP) {
         if ((AreaX >= 0) AND (AreaY >= 0) AND
             (AreaX < width) AND (AreaY < Self->BorderSize)) {
            Self->Direction = MOVE_UP;
            return(EDGE_TOP);
         }
      }

      if (Self->BorderFlags & EDGE_BOTTOM) {
         if ((AreaX >= 0) AND (AreaY >= height - Self->BorderSize) AND
             (AreaX < width) AND (AreaY < height)) {
            Self->Direction = MOVE_DOWN;
            return(EDGE_BOTTOM);
         }
      }

      if (Self->BorderFlags & EDGE_LEFT) {
         if ((AreaX >= 0) AND (AreaY >= 0) AND
             (AreaX < Self->BorderSize) AND (AreaY < height)) {
            Self->Direction = MOVE_LEFT;
            return(EDGE_LEFT);
         }
      }

      if (Self->BorderFlags & EDGE_RIGHT) {
         if ((AreaX >= width - Self->BorderSize) AND (AreaY >= 0) AND
             (AreaX < width) AND (AreaY < height)) {
            Self->Direction = MOVE_RIGHT;
            return(EDGE_RIGHT);
         }
      }

      return(NULL);
   }
   else {
      var.Type = FD_DOUBLE;
      GET_X(Self, &var); x = var.Double;
      GET_Y(Self, &var); y = var.Double;
      GET_Width(Self, &var);  width = var.Double;
      GET_Height(Self, &var); height = var.Double;

      if (AreaX < x) return(NULL);
      else if (AreaY < y) return(NULL);
      else if (AreaX >= x + width) return(NULL);
      else if (AreaY >= y + height) return(NULL);
      else return(-1);
   }
}

/*****************************************************************************
** Resize: ClosingTag
*/

static ERROR RESIZE_ClosingTag(struct Resize *Self, APTR Void)
{
   /*** Check if we are associated with a Render object (this is optional) ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (OwnerID) Self->DrawableID = OwnerID;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Resize: Free
*/

static ERROR RESIZE_Free(struct Resize *Self, APTR Void)
{
   /*** Remove all subscriptions to our Render container ***/

   if (Self->DrawableID) {
      OBJECTPTR Render;
      if (AccessObject(Self->DrawableID, 2000, &Render) IS ERR_Okay) {
         UnsubscribeAction(Render, NULL, GetUniqueID(Self));
         ReleaseObject(Render);
      }
   }

   if (Self->PointerLocked) {
      RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
      Self->PointerLocked = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Resize: Init
*/

static ERROR RESIZE_Init(struct Resize *Self, APTR Void)
{
   OBJECTPTR drawable;
   LONG minwidth, minheight, maxwidth, maxheight;

   if (!Self->DrawableID) Self->DrawableID = GetOwner(Self);

   if (AccessObject(Self->DrawableID, 2000, &drawable) IS ERR_Okay) {
      /* If the drawable has matching dimension restrictions, there is
      ** no point in initialising the resize object.
      */

      if (GetFields(drawable, FID_MinWidth|TLONG,  &minwidth,
                              FID_MinHeight|TLONG, &minheight,
                              FID_MaxWidth|TLONG,  &maxwidth,
                              FID_MaxHeight|TLONG, &maxheight,
                              TAGEND) IS ERR_Okay) {
         if ((minwidth IS maxwidth) AND (minheight IS maxheight)) {
            ReleaseObject(drawable);
            return(ERR_LimitedSuccess);
         }
      }

      SubscribeActionTags(drawable, Self,
         AC_UserClick,
         AC_UserClickRelease,
         AC_UserMovement,
         TAGEND);
      ReleaseObject(drawable);
   }

   /*** If no object was specified for resizing, default to the container ***/

   if (!Self->Object) Self->Object = GetOwner(Self);

   if (Self->BorderFlags) {
      if (Self->BorderFlags & EDGE_TOPLEFT)   Self->Direction = MOVE_UP|MOVE_LEFT;
      else if (Self->BorderFlags & EDGE_TOP)  Self->Direction = MOVE_UP;
      else if (Self->BorderFlags & EDGE_LEFT) Self->Direction = MOVE_LEFT;

      if (Self->BorderFlags & EDGE_TOPRIGHT)    Self->Direction = MOVE_UP|MOVE_RIGHT;
      else if (Self->BorderFlags & EDGE_RIGHT)  Self->Direction = MOVE_RIGHT;
      else if (Self->BorderFlags & EDGE_BOTTOM) Self->Direction = MOVE_DOWN;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Resize: NewObject
*/

static ERROR RESIZE_NewObject(struct Resize *Self, APTR Void)
{
   Self->Button     = 1;
   Self->State      = CLICK_RELEASED;
   Self->Direction  = MOVE_DOWN|MOVE_RIGHT;
   Self->BorderSize = 6.0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Border
Short:  Set this field to define the borders that should be monitored.
Type:   LONG/FLAGS
Prefix: EDGE_
Status: Read/Write

###DESCRIPTION###
<p>If your Resize object needs to monitor the borders of the drawable area
that it is being applied to, use the Border field to specify which borders
should be monitored.  If you do not supply any border flags then the Resize
object will expect you to provide the dimensions for an area to monitor.</p>

<p>Valid Border flags are as follows:</p>

<pre>
   TOP
   BOTTOM
   LEFT
   RIGHT
   TOPLEFT
   TOPRIGHT
   BOTTOMLEFT
   BOTTOMRIGHT
</pre>

<p>The size of the borders that are to be monitored must be set through the
BorderSize field.</p>

###SEE ALSO###
Field: BorderSize

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   BorderSize
Short:  Determines the size of the monitored regions when borders are used.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<P>If you have used the Border field to determine what borders should be
monitored, it is recommended that you set the BorderSize field to define
the size of the border areas.  If you do not set the BorderSize field then
a default value will be used for determining the border size.</p>

###SEE ALSO###
Field: Border

###END###

*****************************************************************************/

static ERROR SET_BorderSize(struct Resize *Self, LONG *Value)
{
   if ((*Value > 0) AND (*Value < 100)) {
      Self->BorderSize = *Value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_SetField, ERR_OutOfRange));
}

/*****************************************************************************

###FIELD###
Name:   Button
Short:  Defines the user button that starts the resize process.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>By default, the user can interact with a resize area by moving the mouse
over it and pressing/holding the left mouse button or its nearest equivalent.
If you would like to change the button that the Resize object reacts to, you
can set the Button field to a different value.  Valid settings are:</p>

<pre>
   1 = Left Mouse Button
   2 = Right Mouse Button
   3 = Middle Mouse Button
</pre>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Direction
Short:  Limits the directions in which the user can apply resizing.
Type:   LONG/FLAGS
Status: Read/Write

###DESCRIPTION###
<p>If you are using a Resize object to monitor a specific region rather than
using the border functionality, you will need to tell the object what
directions the user is allowed to apply the resize.  Valid directions are
UP, DOWN, LEFT and RIGHT.  Setting a direction such as UP|LEFT would allow
the user to resize towards the top left corner of the display, but not the
bottom right corner.</p>

<p>If you have set the Border field then there is no need to set the Direction
as it will be ignored.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Specifies the height of the resize area.
Type:   FLOAT|PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>A resize area can be given a fixed or relative height by setting this
field to the desired value.  To set a relative height, use the FD_PERCENT
flag when setting the field - otherwise a fixed height is assumed.</p>

###SEE ALSO###
Fields: Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_Height(struct Resize *Self, struct Variable *Value)
{
   OBJECTPTR Render;
   FLOAT ContainerHeight, YCoord, YOffset;

   if (Value->Type IS FD_DOUBLE) {
      if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->Height * ContainerHeight / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetFields(Render, FID_Height|TFLOAT,  &ContainerHeight,
                              FID_YCoord|TFLOAT,  &YCoord,
                              FID_YOffset|TFLOAT, &YOffset,
                              TAGEND);
            Value->Double = ContainerHeight - YOffset - YCoord;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_FieldNotSet);
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->Height / ContainerHeight * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FD_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_Y) YCoord = Self->YCoord;
            else YCoord = Self->YCoord * 100 / ContainerHeight;
            if (Self->Dimensions & DMF_FIXED_YOFFSET) YOffset = Self->YOffset;
            else YOffset = Self->YOffset * 100 / ContainerHeight;
            Value->Double = (ContainerHeight - YOffset - YCoord) / ContainerHeight * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_FieldNotSet);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("!SetField:","[Gradient] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
      Self->Height     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("!SetField:","[Gradient] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
      Self->Height     = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Object
Short:  Defines the object that is to be the recipient of the Resize() action.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>This field determines the object that receives resize messages when the
user interacts with the Resize object.  By default the Resize object's
container will receive the messages, but setting this field directly allows
you to change who the recipient is.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Specifies the width of the resize area.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>A resize area can be given a fixed or relative width by setting this
field to the desired value.  To set a relative width, use the FD_PERCENT
flag when setting the field - otherwise a fixed width is assumed.</p>

###SEE ALSO###
Fields: Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_Width(struct Resize *Self, struct Variable *Value)
{
   OBJECTPTR Render;
   FLOAT ContainerWidth, XCoord, XOffset;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width * ContainerWidth / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = ContainerWidth * Self->XCoord / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - XCoord - Self->XOffset;
            else Value->Double = ContainerWidth - XCoord - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_FIXED_WIDTH) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width / ContainerWidth * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FD_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = Self->XCoord * 100 / ContainerWidth;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) XOffset = Self->XOffset;
            else XOffset = Self->XOffset * 100 / ContainerWidth;
            Value->Double = (ContainerWidth - XOffset - XCoord) * 100 / ContainerWidth;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_FieldNotSet);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_Width(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
      Self->Width      = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
      Self->Width      = Value->Double;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  Specifies the horizontal beginning of the resize area.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The horizontal position of a resize area can be set to an absolute or
relative coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      OBJECTPTR Render;
      FLOAT ContainerWidth, Width;
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_X) Value->Double = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->XCoord * ContainerWidth / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_WIDTH) Width = Self->Width;
            else Width = ContainerWidth * Self->Width / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - Width - Self->XOffset;
            else Value->Double = ContainerWidth - Width - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
      Self->XCoord     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
      Self->XCoord     = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Specifies the horizontal offset of the resize area.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the resize area will be
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the area is dynamically
calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the resize area will start from an X coordinate calculated from the
formula "XCoord = ContainerWidth - ResizeWidth - XOffset".</p>

###SEE ALSO###
Field: Width, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      OBJECTPTR Render;
      FLOAT ContainerWidth, Width;

      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_XOFFSET) {
         Value->Double = Self->XOffset;
      }
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->XOffset * ContainerWidth / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_WIDTH) Width = Self->Width;
            else Width = ContainerWidth * Self->Width / 100;
            if (Self->Dimensions & DMF_FIXED_X) Value->Double = ContainerWidth - (Self->XCoord + Width);
            else Value->Double = ContainerWidth - ((Self->XCoord * ContainerWidth / 100) + Width);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
      Self->XOffset    = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
      Self->XOffset    = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  Specifies the vertical beginning of the resize area.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The vertical position of a resize area can be set to an absolute or
relative coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_Y(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      OBJECTPTR Render;
      FLOAT ContainerHeight, Height;

      if (Self->Dimensions & DMF_FIXED_Y) {
         Value->Double = Self->YCoord;
      }
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->YCoord * ContainerHeight / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_HEIGHT) Height = Self->Height;
            else Height = ContainerHeight * Self->Height / 100;
            if (Self->Dimensions & DMF_FIXED_YOFFSET) Value->Double = ContainerHeight - Height - Self->YOffset;
            else Value->Double = ContainerHeight - Height - (ContainerHeight * Self->YOffset / 100);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {

   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
      Self->YCoord     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
      Self->YCoord     = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Specifies the vertical offset of the resize area.
Type:   FLOAT
Status: Write

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the resize area will be
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the area is dynamically
calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the resize area will start from a Y coordinate calculated from the
formula "YCoord = ContainerHeight - ResizeHeight - YOffset".</p>

###SEE ALSO###
Field: Height, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      OBJECTPTR Drawable;
      FLOAT ContainerHeight, Height;

      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_YOFFSET) {
         Value->Double = Self->YOffset;
      }
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->YOffset * ContainerHeight / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_HEIGHT) Height = Self->Height;
            else Height = ContainerHeight * Self->Height / 100;
            if (Self->Dimensions & DMF_FIXED_Y) Value->Double = ContainerHeight - (Self->YCoord + Height);
            else Value->Double = ContainerHeight - ((Self->YCoord * ContainerHeight / 100) + Height);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Resize *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
      Self->YOffset    = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
      Self->YOffset    = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: GetSystemPointer()
*/

static ERROR GetSystemPointer(struct Pointer **Pointer)
{
   ERROR error;
   OBJECTPTR object;

   if ((error = AccessObject(glPointerID, 2000, (OBJECTPTR *)Pointer)) != ERR_Okay) {
      if (error IS ERR_NoMatchingObject) {
         if (FastFindObject("SystemPointer", ID_POINTER, &glPointerID, 1, NULL) != ERR_Okay) return(ERR_Search);
         if (AccessObject(glPointerID, 2000, &object) IS ERR_Okay) {
            *Pointer = (struct Pointer *)object;
         }
         else {
            DPrintF("@GetSystemPointer()","[Resize} Failed to access the mouse pointer.");
            return(ERR_AccessObject);
         }
      }
      else return(error);
   }

   return(ERR_Okay);
}

/*****************************************************************************

                         "Come to the Sabbath
                          Down by the ruined bridge
                          Later on the master will join us
                          Called from the heart of Hell"

                               - King Diamond

*****************************************************************************/

