/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Keymaker
Version:   1.0
ID:        ID_KEYMAKER
Category:  Command/Tool
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2001-2004.  All rights reserved.
Keywords:  keymap, key, map, mapper, keyboard, international
Short:     Generates keymap files from user defined configuration

###DESCRIPTION###

###ACTIONS###
Activate:   Runs the keymap maker and prints the results.
ClosingTag: If the object is non-static, this action will run the keymap maker process and then the object will self-destruct.

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/files/all.h>
#include <pandora/data/config.h>
#include <pandora/input/keyboard.h>
#include <pandora/modules/strings.h>
#include <pandora/input/keymaptable.h>
#define UNICODE_TABLE TRUE
#include <pandora/input/unicodetable.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_KEYMAKER 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_KEYMAKER, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2001.  All rights reserved.",
   "January 2004",
   "Keymaker"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR KeymakerClass = 0;
static OBJECTPTR StringsMod = 0;

/*****************************************************************************
** Class definition.
*/

struct Keymaker {
   OBJECT_HEADER             /* Class header */
   OBJECTID OutputID;
   LONG Static;
   STRING Location;       /* Location of the document directory or specific file */

   PRIVATE_FIELDS
};

static ERROR GET_Location(struct Keymaker *, STRING *);

static ERROR SET_Location(struct Keymaker *, STRING);

static struct FieldArray KeymakerFields[] = {
   { "Output", 0, FDF_OBJECTID|FDF_RW, 0, NULL, NULL },
   { "Static", 0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   /*** Virtual fields ***/
   { "Location", 0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   { "Src",      0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   END_FIELD
};

/*****************************************************************************
** Class actions.
*/

static ERROR KEYMAKER_Activate(struct Keymaker *, APTR);
static ERROR KEYMAKER_ClosingTag(struct Keymaker *, APTR);
static ERROR KEYMAKER_Free(struct Keymaker *, APTR);

static struct ActionArray KeymakerActions[] = {
   { AC_Activate,   KEYMAKER_Activate },
   { AC_ClosingTag, KEYMAKER_ClosingTag },
   { AC_Free,       KEYMAKER_Free },
   { NULL, NULL }
};

/*****************************************************************************
** Class methods.
*/

#define MTK_About -1

static ERROR KEYMAKER_About(struct Keymaker *, APTR);

static struct MethodArray KeymakerMethods[] = {
   { MTK_About,  KEYMAKER_About,  "About",  NULL },
   { NULL, NULL, NULL, NULL }
};

static UBYTE About[] = {
"Key Maker\n\
\n\
Copyright Rocklyte Systems (c) 2001-2004.  All rights reserved.\n\
\n\
This program builds keymap files that can be used to match the characters\n\
on international keyboards.  You can also use it to bind special key\n\
symbols to your keyboard if you require access to certain Unicode\n\
characters.  To use the Key Maker, you simply need to provide it with\n\
a keymap configuration file.  The resulting keymap file will be written\n\
to the system:config/keymaps/ directory, ready for use.  Your current\n\
keymap can be changed by altering the user:config/keyboard.cfg file.\n"
};

/*****************************************************************************
** Function Prototypes
*/

static void OutputLine(struct Keymaker *, STRING);
static void OutputText(struct Keymaker *, STRING);

/*****************************************************************************
** Stubs.
*/

static void WriteWord(OBJECTPTR File, WORD Value)
{
   struct mtWriteFileWord args;
   args.Value = Value;
   Action(MT_WriteFileWord, File, &args);
}

static void WriteLong(OBJECTPTR File, LONG Value)
{
   struct mtWriteFileLong args;
   args.Value = Value;
   Action(MT_WriteFileLong, File, &args);
}

static LONG Write(OBJECTPTR File, APTR Buffer, LONG Size)
{
   struct acWrite write;
   write.Buffer = Buffer;
   write.Length = Size;
   if (Action(AC_Write, File, &write) IS ERR_Okay) return(write.Result);
   else return(NULL);
}

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&KeymakerClass, NULL,
      FID_BaseClassID|TLONG, ID_KEYMAKER,
      FID_Version|TFLOAT,    VER_KEYMAKER,
      FID_Name|TSTRING,      "Keymaker",
      FID_Category|TLONG,    CCF_TOOL,
      FID_Actions|TPTR,      KeymakerActions,
      FID_Methods|TPTR,      KeymakerMethods,
      FID_Fields|TPTR,       KeymakerFields,
      FID_Size|TLONG,        sizeof(struct Keymaker),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (StringsMod)    { Action(AC_Free, StringsMod, NULL); StringsMod = NULL; }
   if (KeymakerClass) { Action(AC_Free, KeymakerClass, NULL); KeymakerClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  About
Short: Prints summarised information about the Keymaker class.

###DESCRIPTION###
<p>If this method is called the Keymaker object will print a summary of
what the Keymaker class does, who wrote it, copyright information and so
on.</p>

###RESULT###
ERR_Okay: The method executed successfully.
ERR_FieldNotSet: The Output field has not been set.

###END###

*****************************************************************************/

static ERROR KEYMAKER_About(struct Keymaker *Self, APTR Void)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      ActionMsg(AC_Clear, Self->OutputID, NULL);

      channel.ObjectID     = GetUniqueID(Self);
      channel.DataType     = DATA_TEXT;
      channel.Version      = 1;
      channel.Buffer       = About;
      channel.Size         = StrLength(About) + 1;
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, MTK_About, ERR_FieldNotSet));
}

/*****************************************************************************
** Keymaker: Activate
*/

static ERROR KEYMAKER_Activate(struct Keymaker *Self, APTR Void)
{
   #define MAX_KEYS 6
   struct ConfigEntry *entries;
   struct mtReadConfig read;
   OBJECTPTR config, file;
   STRING str, destfile;
   LONG i, j, k, amtentries, total_keys;
   WORD keycode;
   UBYTE KeymapName[KEYMAP_NAMELEN], buffer[300], unicode_name[100];
   UWORD unicode;

   if (!Self->Location) return(ObjectError(Self, ERH_Activate, ERR_FieldNotSet));

   DPrintF("~Activate()","[Keymaker:%d] Generating keymap from %s", Self->Head.UniqueID, Self->Location);

   if (CreateObject(ID_CONFIG, NULL, &config, NULL,
      FID_Location, Self->Location,
      TAGEND) IS ERR_Okay) {

      if (GetFields(config, FID_Entries|TPTR,     &entries,
                            FID_AmtEntries|TLONG, &amtentries,
                            TAGEND) IS ERR_Okay) {

         /*** Retrieve the name of the keymap and use it to build the destination file ***/

         read.Section = "INFO";
         read.Item = "Name";
         Action(MT_ReadConfig, config, &read);
         destfile = read.Result;

         StrFormat(buffer, sizeof(buffer), "system:config/keymaps/%s", destfile);

         if (CreateObject(ID_FILE, NULL, &file, NULL,
            FID_Location|TSTRING, buffer,
            FID_Flags|TLONG,      FL_WRITE|FL_NEW,
            TAGEND) IS ERR_Okay) {

            if (Self->OutputID) {
               ActionMsg(AC_Clear, Self->OutputID, NULL);
            }

            /*** Write out the main header ***/

            Write(file, "KEYMAP10", 8);
            WriteLong(file, sizeof(struct KeymapHeader));
            WriteLong(file, amtentries);
            WriteLong(file, MAX_KEYS+1);
            Write(file, KeymapName, KEYMAP_NAMELEN);

            /*** Write out the key values ***/

            for (i=0; (i < amtentries) AND (StrCompare("LAYOUT", entries[i].Section, 0, STR_MATCHLENGTH) != ERR_Okay); i++);

            total_keys = 0;
            for (; i < amtentries; i++) {
               for (keycode=0; keycode < K_LIST_END; keycode++) {
                  if (StrCompare(glKeymapTable[keycode], entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {

                     WriteWord(file, keycode); /* Key code */

                     /*** Write out the series of unicode values assigned to this key ***/

                     str = entries[i].Data;
                     for (j=0; (j < MAX_KEYS); j++) {
                        unicode = NULL;

                        if (*str) {
                           while ((*str) AND (*str <= 0x20)) str++;
                           for (k=0; (*str) AND (*str != ','); k++) unicode_name[k] = *str++;
                           if (*str IS ',') str++;
                           unicode_name[k] = 0;

                           if (k > 0) {
                              for (k=0; glUnicodeTable[k].Name; k++) {
                                 if (StrCompare(unicode_name, glUnicodeTable[k].Name, 0, STR_MATCHLENGTH|STR_MATCHCASE) IS ERR_Okay) {
                                    unicode = glUnicodeTable[k].Value;
                                    break;
                                 }
                              }

                              if (!unicode) {
                                 StrFormat(buffer, sizeof(buffer), "Unicode character \"%s\" not recognised.", unicode_name);
                                 OutputLine(Self, buffer);
                              }
                           }
                        }

                        WriteWord(file, unicode); /* Unicode character.  NULL must still be used as a placeholder if the key was invalid */
                     }
                     total_keys++;
                     break;
                  }
               }

               if (keycode >= K_LIST_END) {
                  StrFormat(buffer, sizeof(buffer), "Key code \"%s\" is not a recognised key code.", entries[i].Item);
                  OutputLine(Self, buffer);
               }
            }

            /*** Send process information to the output object ***/

            if (Self->OutputID) {
               StrFormat(buffer, sizeof(buffer), "\n%d keys were found in the keymap definition.", total_keys);
               OutputLine(Self, buffer);

               StrFormat(buffer, sizeof(buffer), "\nKeymap written to \"system:config/keymaps/%s\"", destfile);
               OutputLine(Self, buffer);
            }

            acFree(file);
         }
         else OutputLine(Self, "Failed to open the destination file.");
      }
      else OutputLine(Self, "Source keymap contains no information.");

      acFree(config);
   }
   else OutputLine(Self, "Failed to load the keymap configuration file.");

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Keymaker: ClosingTag
*/

static ERROR KEYMAKER_ClosingTag(struct Keymaker *Self, APTR Void)
{
   OutputText(Self, "Keymap Generator\n\nCopyright Rocklyte Systems (c) 2001-2002.  All rights reserved.\n");

   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
      Action(AC_Free, Self, NULL);
      return(ERR_Okay);
   }
   else return(ERR_Okay);
}

/*****************************************************************************
** Keymaker: Free
*/

static ERROR KEYMAKER_Free(struct Keymaker *Self, APTR Void)
{
   if (Self->Location) {
      FreeMemory(Self->Location);
      Self->Location = NULL;
   }

   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    Must refer to the location of a keymap configuration file.
Type:     STRING
Status:   Read/Write

###DESCRIPTION###
<p>Before a Keymaker object can be activated, you must give it a configuration
file to process.  This requires that you set the Location field so that it
refers to a keymap configuration file.</p>

###END###

****************************************************************************/

static ERROR GET_Location(struct Keymaker *Self, STRING *Value)
{
   if (Self->Location) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   *Value = NULL;
   return(ERR_FieldNotSet);
}

static ERROR SET_Location(struct Keymaker *Self, STRING Value)
{
   LONG i;

   if (Self->Location)  { FreeMemory(Self->Location); Self->Location = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->Location, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
         Self->Location[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Output
Short:  Set this field to give the Keymaker an object to send output to.
Status: Read/Write
Type:   OBJECTID

###DESCRIPTION###
<p>The Output field must be set in order for a Keymaker object to
process the commands sent from the user.  If it is not set then commands
will not be executed when the Keymaker object is Activated.  The Output
object must accept text-based information via data channels in order for the
data to be interpreted correctly.  If the Output object does not support
text based information, the Keymaker will still process the command but
the user will not receive any text based output.</p>

<p>It is recommended that this field points to an object that is a member
of the Text class.</p>

###SEE ALSO###
Class: Text

###END###

*****************************************************************************/

/****************************************************************************/

static void OutputText(struct Keymaker *Self, STRING Buffer)
{
   struct acDataChannel Channel;

   if (Self->OutputID) {
      /*** Send the data to the output object ***/

      ActionMsg(AC_Clear, Self->OutputID, NULL);

      Channel.ObjectID     = GetUniqueID(Self);
      Channel.DataType     = DATA_TEXT;
      Channel.Version      = 1;
      Channel.Buffer       = Buffer;
      for (Channel.Size=0; Buffer[Channel.Size]; Channel.Size++);
      Channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &Channel);
   }
}

static void OutputLine(struct Keymaker *Self, STRING Buffer)
{
   struct acDataChannel Channel;

   if (Self->OutputID) {
      /*** Send the data to the output object ***/
      Channel.ObjectID     = GetUniqueID(Self);
      Channel.DataType     = DATA_TEXT;
      Channel.Version      = 1;
      Channel.Buffer       = Buffer;
      for (Channel.Size=0; Buffer[Channel.Size]; Channel.Size++);
      Channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &Channel);
   }
}
